#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import subprocess
import paramiko
import sys
import re
import socket
import threading
import gobject
gobject.threads_init()

from xdg.BaseDirectory import *
from ConfigParser import SafeConfigParser

from tacix.util import *
from tacix.freenx.FreeNXShell import FreeNXShell

GENERIC_LINE = re.compile(r'^NX> (?P<protnr>\d+) (?P<protcmd>.*)')
PROPERTIES = { '700': ('Session id', 'id'),
        '705': ('Session display', 'display'),
        '703': ('Session type', 'type'),
        '701': ('Proxy cookie', 'proxy_cookie'),
        '702': ('Proxy IP', 'proxyip'),
        '706': ('Agent cookie', 'cookie'),
        '704': ('Session cache', 'type'),
        '707': ('SSL tunneling', 'encryption')}

STATUS = ['1009', '710']
IGNORED = ['716', '1002', '105']

DEFAULT_OPTIONS = {'link':'modem',
        'backingstore': '1',
        'encryption': '1',
        'cache': '16M',
        'images': '64M',
        'shmem': '1',
        'shpix': '1',
        'strict': '0',
        'composite': '1',
        'media': '0',
        'session': 'localhost',
        'type': 'unix-gnome',
        'geometry': '1440x850',
        'client': 'linux',
        'keyboard': 'pc101/us',
        'screeninfo': '1440x850x24+render'}

class TXClient(FreeNXShell):
    def __init__(self, host='localhost', port=22, session_name=None):
        self.options = {}
        self.options.update(DEFAULT_OPTIONS)
        self.options['geometry'] = self.get_geometry()
        self.session_name = session_name
        self.options.update(self.get_config_dict())
        
        self.txpath = None
        self.host = host
        self.port = port

        self.logger = get_home_logger('txclient')
        #self.logger = None

        self.resumablesessions = None
        self.stage = None
        self.channels = None
        self.nxproxy_proc = None

        self.connect_callback = None
        self.started_signal_callback = None
        self.disconnect_callback = None
        self.update_status_callback = None

        self.loop = None

    # Options manipulations functions
    def write_options(self):
        opts = ['nx/nx']
        self.debug('Creating options file')

        for opt_name in ("keyboard", "id", "clipboard", "geometry",
                "option", "shadow", "shadowmode", "streaming", "defer", "tile",
                "R", "render", "fullscreen", "shpix", "shmem", "encryption",
                "resize", "backingstore", "menu", "shadowuid", "clients",
                "client", "copysize", "cookie", "port", "connect",
                # Server only properties
                # "cache", "images", "type", "link"
                ):
            if opt_name in self.options and self.options[opt_name] != '':
                opts.append('%s=%s' % (opt_name, self.options[opt_name]))

        self.debug('options file created')
        open( self.txpath.get_filepath( "options" ), "w").write("%s:%s\n" % (",".join(opts), self.options['display']))

    def get_geometry(self):
        xprop = subprocess.Popen(["/usr/bin/xprop", "-root", "_NET_WORKAREA"], stdout=subprocess.PIPE)
        line = xprop.stdout.read(1024)
        xprop.wait()
        line = line.strip()
        if not '=' in line:
            return "800x600"
        xprops = line.split('=')[1].split(',')
        width = xprops[2].strip()
        height = str(int(xprops[3].strip()) -25)
        return width + 'x' + height

    # Protocol handler
    def handle_line(self, line):
        if line is None: 
            return True
        line = line.strip()
        if line == '':
            return True

        self.debug(line)
        match = GENERIC_LINE.search(line)
        if match is None:
            if self.stage is not None and self.stage == 'listsession':
                return self.add_running_session(line)
            if line == 'NX> 105':
                return True
            self.debug('Could not parse line %s' % line)
            return True
        protnr = match.group('protnr')
        protcmd = match.group('protcmd')
        self.debug(protnr)
        self.debug(protcmd)
        if protnr in IGNORED:
            return True
        if protnr in STATUS:
            return True
        if protnr in PROPERTIES:
            prop, key = PROPERTIES[protnr]
            value = protcmd.split(':')[1].strip()
            self.options[key] = value
            self.debug('Added value %s to key %s' % (value, key))
            return True
        handler_name = '_%s_handler' % protnr
        try:
            handler_method = getattr(self, handler_name)
            self.debug('handler %s found' % handler_name)
            return handler_method(line)
        except AttributeError, e:
            self.debug('Unhandled nx command %r (%s)' % (protnr, e))

        return True

    def _1000_handler(self, line):
        self.stdin.write('set_tacix_mode')
        self.stdin.write('\n')
        self.stdin.write('listsession --user="%(user)s" --type="%(type)s"' % self.options)
        self.stdin.write('\n')
        self.stdin.flush()
        return True

    def _127_handler(self, line):
        self.stage = 'listsession'
        return True

    def _148_handler(self, line):
        self.stage = None
        if self.resumablesessions is not None:
            for session in self.resumablesessions:
                if session['state'] != 'terminated':
                    self.options['id'] = self.resumablesessions[0]['uniqueid']
        cmd = []
        for property in self.options.keys():
            cmd.append('--%s="%s"' % (property, self.options[property]))
        send = None
        if 'id' in self.options:
            send = 'restoresession %s' % ' '.join(cmd)
        else:
            send = 'startsession %s' % ' '.join(cmd)
        self.stdin.write(send + '\n')
        self.debug(send)
        self.stdin.flush()
        return True

    def _1001_handler(self, line):
        return self.quit()

    def _999_handler(self, line):
        return True

    def _1005_handler(self, line):
        status_matcher = re.compile(r'^NX> (?P<protcmd>.*): (?P<status>.*)')
        match = status_matcher.search(line)
        status = match.group('status').strip()
        self.debug('Status is now %s' % status)
        if self.update_status_callback is not None:
            self.update_status_callback(status)
        if status == 'running' and self.started_signal_callback is not None:
            self.debug('Session started')
            self.started_signal_callback()
        if status == 'terminated':
            self.quit()
            return False
        return True

    def _1006_handler(self, line):
        self.stdin.write('bye')
        self.stdin.write('\n')
        self.stdin.flush()
        self.run_proxy()
        return True

    def quit(self):
        self.debug('quit solicited')
        #if self.nxproxy_proc is not None:
        #    self.nxproxy_proc.wait()
        self.debug('nxproc exited')
        if self.channels is not None:
            for chan in self.channels:
                chan.close()
        self.debug('channels closed')
        #if self.thread_tunnel is not None:
        #    self.thread_tunnel.join()
        #self.debug('tunnel thread joined')
        if self.disconnect_callback is not None:
            self.disconnect_callback()
        self.debug('disconnect emmited')
        if self.loop is not None:
            self.loop.quit()
        else:
            self.debug('Loop is none')
        return False

    def add_running_session(self, line):
        options = {}
        options.update({
                  'display': line[:7].strip(),
                  'type': line[8:24].strip(),
                  'uniqueid': line[25:57].strip(),
                  'cli_opts': line[58:66].strip(),
                  'depth': line[67:72].strip(),
                  'screen': line[73:87].strip(),
                  'state': line[88:99].strip(),
                  'session': line[100:].strip(),
                  })
        if options['display'] == 'Display' or options['display'] == '-------':
            return True
        self.debug(options)
        if self.resumablesessions is None:
            self.resumablesessions = [options]
        else:
            self.resumablesessions.append(options)
        return True

    # Actions
    def connect(self, user, password):
        self.options['user'] = user
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            self.client.connect(self.host, port=self.port, username=user, password=password, allow_agent=False, look_for_keys=False, timeout=10)
        except paramiko.AuthenticationException, e:
            return False
        command = '/usr/lib/tacix/tacix-freenx-node --slave'
        self.chan = self.client.get_transport().open_session()
        self.chan.exec_command(command)
        self.chan.settimeout(0.0)
        self.stdin = self.chan.makefile('wb', -1)
        self.stdout = self.chan.makefile('rb', -1)
        return True

    def debug(self, log):
        if self.logger is not None:
            self.logger.debug(log)
        else:
            print(log)

    def run(self, connect_callback=None, disconnect_callback=None, started_signal_callback=None, update_status_callback=None):
        self.started_signal_callback = started_signal_callback
        self.connect_callback = connect_callback
        self.disconnect_callback = disconnect_callback
        self.update_status_callback = update_status_callback
        gobject.io_add_watch(self.chan, gobject.IO_IN | gobject.IO_HUP, self.read_line)

    def wait_loop(self):
        self.loop = gobject.MainLoop()
        self.loop.run()
        self.debug('loop exited')

    def read_line(self, source, condition):
        if condition == gobject.IO_HUP:
            self.quit()
            return False
        ret = False
        try:
            x = self.chan.recv(1024)
            self.debug(x)
            ret = self._receive(x, self.handle_line)
        except Exception, e:
            pass
        return ret

    def run_proxy(self):
        self.txpath = TXPath( self.options['id'], create=True, client=True )
        display = self.options['display']

        os.environ['NX_DISPLAY'] = 'nx/nx,options=%s:%s' % (self.txpath.get_filepath('options'), self.options['display'])
        command = []
        self.chan.settimeout(None)

        command.append('/usr/bin/nxproxy')
        command.append('-S')
        if self.options['encryption'] == '1':
            remote_display_port = int(display) + 4000
            aux = 0
            self.debug('Setting up tunnel')
            local_port = None
            import random
            while True:
                try:
                    local_port = random.randint(1,1000) + 15000
                    self.thread_tunnel = threading.Thread(target=self.tunnel_remote_port, args=(remote_display_port, local_port))
                    self.thread_tunnel.start()
                    break
                except Exception, e:
                    self.debug('%s' % e)
                    aux = aux + 1
                    if aux == 3:
                        sys.exit(1)

            self.options['port'] = str(local_port)
            self.options['connect'] = 'localhost'
            self.write_options( )
            self.debug('Tunnel created')
        else:
            self.write_options( )
            command.append('%s:%s' % (self.host, display))
        self.nxproxy_proc = subprocess.Popen(command, close_fds=True)

    def pulseaudio(self):
        '''
           pulseaudio --use-pid-file=false --file pulseaudio.conf -n
           pulseaudio.conf:
           load-module module-esound-protocol-tcp auth-anonymous=1
           load-module module-native-protocol-tcp auth-anonymous=1

           server-side:
           PULSE_SERVER=localhost:4000
        '''
        return

    def tunnel_remote_port(self, remote_port, local_port):
        HOST = ''                 # Symbolic name meaning all available interfaces
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((HOST, local_port))
        s.listen(1)
        conn, addr = s.accept()
        print 'Connected by', addr
        chan = None
        try:
            chan = self.client.get_transport().open_channel('direct-tcpip',
                                                   ('127.0.0.1', remote_port),
                                                   conn.getpeername())
        except Exception, e:
            raise Exception('Incoming request to %s:%d failed: %s' %
                    ('127.0.0.1', remote_port, repr(e)))
        if chan is None:
            raise Exception('Incoming request to %s:%d was rejected by the SSH server.' %
                    ('127.0.0.1', remote_port))

        gobject.io_add_watch(chan, gobject.IO_IN | gobject.IO_HUP, self.forward_input, conn)
        gobject.io_add_watch(conn, gobject.IO_IN | gobject.IO_HUP, self.forward_input, chan)
        self.channels = [chan, conn]

    def forward_input(self, source, condition, dest):
        if condition == gobject.IO_HUP:
            self.debug('Connect terminated')
            print('Connect terminated' + str(source))
            self.quit()
            return False
        source.settimeout(0.0)
        data = source.recv(4096)
        if len(data) == 0:
            self.debug('Connect terminated 2')
            print('Connect terminated 2' + str(source))
            self.quit()
            return False
        dest.send(data)
        source.settimeout(None)
        return True

    def save_config_file(self):
        if self.session_name is None:
            return
        filename = save_config_path("tacix") + "/" + self.session_name + ".tx"
        parser = SafeConfigParser()
        if not parser.has_section(self.session_name):
            parser.add_section(self.session_name)
        for prop in self.options:
            parser.set(self.session_name, prop, self.options[prop])
        file = open(filename, "w")
        parser.write(file)
        file.close()

    def get_config_dict(self):
        dict = {}
        if self.session_name is None:
            return dict
        filename = load_first_config('tacix', self.session_name + ".tx")
        if filename is None:
            return dict
        parser = SafeConfigParser()
        try:
            parser.read(filename)
        except Exception, e:
            pass
        try:
            if not parser.has_section(self.session_name):
                return dict
            for opt in parser.options(self.session_name):
                opts = parser.get(self.session_name, opt).split(' ')
                if len(opts) > 1:
                    dict[opt] = opts
                else:
                    dict[opt] = opts[0]
        except Exception,e:
            pass
        return dict

