#!/usr/bin/python
#
# Copyright (C) 2006 - Florian Ludwig <dino@phidev.info>
#                      Franz Pletz <fpletz@franz-pletz.org>
#               2009 - Marcelo Boveto Shima <marceloshima@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)




import os
import logging
import socket

import avahi
import dbus, dbus.glib
import gobject

SERVICE_TYPE = '_tacix._tcp'
DOMAIN = ''
host = ''

class AvahiService:
    def __init__(self):
        self.group = None
        self.name_counter = 0

        self.name = "TaciX server on %s" % socket.gethostname()
        self.host = host
        self.port = '22'

    def _remove_service(self):
        if self.group is not None:
            self.group.Reset()

    def _add_service(self):
        if self.group is None:
            self.group = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME,
                    self.server.EntryGroupNew()),
                    avahi.DBUS_INTERFACE_ENTRY_GROUP)
            self.group.connect_to_signal('StateChanged',
                    self._group_state_changed)
        else:
            return

        assert self.group.IsEmpty()
        logging.debug('Creating service \'%s\' (%s)' %
                (self.name, SERVICE_TYPE))

        self.group.AddService(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, 0,
                self.name, SERVICE_TYPE, DOMAIN, host,
                dbus.UInt16(self.port),
                avahi.string_array_to_txt_array([]))
        self.group.Commit()

    def _group_state_changed(self, state, wtf):
        if state == avahi.ENTRY_GROUP_ESTABLISHED:
            logging.debug('Service established')
        elif state == avahi.ENTRY_GROUP_COLLISION:
            self.name_counter += 1
            if self.name_counter >= 12:
                logging.error('No suitable service name found after %i '
                        'retries, exiting' % self.name_counter)
                self.stop()
            else:
                self.name = server.GetAlternativeServiceName(name)
                logging.warning('Service name collision, changing name to '
                        '\'%s\'' % name)
                self._remove_service()
                self._add_service()

    def _server_state_changed(self, state, *args):
        if state == avahi.SERVER_COLLISION:
            logging.warning('Server name collision')
            self._remove_service()
        elif state == avahi.SERVER_RUNNING:
            self._add_service()

    def _name_owner_changed(self, name, old, new):
        if name == avahi.DBUS_NAME:
            if new != '':
                logging.info('Avahi server appeared')
                self.lock.release()
                self._avahi_connect()
                self.lock.acquire()
            elif old != '':
                logging.debug('Avahi server disappeard')
                self.group = None

    def _dbus_connect(self):
        self.dbus = dbus.Interface(
                self.bus.get_object('org.freedesktop.DBus',
                    '/org/freedesktop/DBus'), 'org.freedesktop.DBus')
        self.dbus.connect_to_signal('NameOwnerChanged', self._name_owner_changed)

    def _avahi_connect(self):
        try:
            self.server = dbus.Interface(
                    self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER),
                    avahi.DBUS_INTERFACE_SERVER)
        except dbus.DBusException:
            return

        self.server.connect_to_signal('StateChanged', self._server_state_changed)
        self._server_state_changed(self.server.GetState())

        self.service_browser = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME,
            self.server.ServiceBrowserNew(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC,
                SERVICE_TYPE, 'local', dbus.UInt32(0))),
            avahi.DBUS_INTERFACE_SERVICE_BROWSER)

    def add_itemnew_signal(self, func):
        self.service_browser.connect_to_signal('ItemNew', func)

    def add_itemremove_signal(self, func):
        self.service_browser.connect_to_signal('ItemRemove', func)

    def setup(self):
        self.bus = dbus.SystemBus()
        self._dbus_connect()
        self._avahi_connect()

    def quit(self):
        if self.group is not None:
            self.group.Free()

    def run(self):
        self.loop()
        self.quit()

    def loop(self):
        self.ml = gobject.MainLoop()
        self.ml.run()

    def stop(self):
        self.ml.quit()


    ''' Test methods '''
    def setup_test_signals(self):
        self.add_itemnew_signal(self.service_added)
        self.add_itemremove_signal(self.service_removed)

    def service_added(self, interface, protocol, name, type, domain, flags):
        print ('Added tacix server: interface = %s, protocol = %s, name = %s, type = %s, domain = %s, flags = %s' %
               (interface, protocol, name, type, domain, flags))
        #self._remove_service()

        #self['%s:%s:%s' % (name, interface, domain)] = {}
        self.server.ResolveService(int(interface), int(protocol),
                name, type, domain, avahi.PROTO_UNSPEC, dbus.UInt32(0),
                reply_handler=self.service_resolved,
                error_handler=logging.error)

    def service_resolved(self, interface, protocol, name, type, domain, host, aprotocol, address, port, txt, flags):
        print ('Resolved tacix server: interface = %s, protocol = %s, name = %s, type = %s, domain = %s, flags = %s' %
               (interface, protocol, name, type, domain, flags))
        if address.startswith('fe80:'):
            logging.info('%s -> %s: Local Link IPv6 addresses not supported!'
                % (host, address))
            return

        url = '%s:%s' % (address, port)
        print url
        #self['%s:%s:%s' % (name, interface, domain)] = {
        #        'host': host, 'address': address, 'port': port, 'txt': txt,
        #        'proto': protocol, 'url': url,
        #        'pkglist': self._fetch_list(url)}
        logging.debug('New host \'%s\' (%s)' %
                (name, url))


    def service_removed(self, interface, protocol, name, type, domain, flags):
        print ('Removed tacix server: interface = %s, protocol = %s, name = %s, type = %s, domain = %s, flags = %s' %
               (interface, protocol, name, type, domain, flags))

