#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import dbus
import dbus.service

import os
import sys
import gobject
import time
import subprocess

import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

from session import Session
from tacix import util
from tacix.server.nx import provider
from tacix.util import *

logger = util.get_logger('txmanager', '/var/log/tacixmanager.log')

class SessionManager( dbus.service.Object ):
    def __init__( self ):
        logger.debug("Starting")

        self.loop = gobject.MainLoop()

        # Getting the system dbus
        self.bus = dbus.SystemBus( )
        
        bus_name = dbus.service.BusName(BUS_NAME, self.bus)
        dbus.service.Object.__init__(self, bus_name=bus_name, object_path=MANAGER_PATH)

        # Getting the ConsoleKit object
        self.ck_manager = dbus.Interface (
                self.bus.get_object ('org.freedesktop.ConsoleKit', '/org/freedesktop/ConsoleKit/Manager'),
                'org.freedesktop.ConsoleKit.Manager')

        # Dicts for store the vars
        self.manager_sessions = {}
        self.printers = {}
        self.mounts = {}
        self.managed_seats = {}

        self.dbus_info = dbus.Interface(self.bus.get_object('org.freedesktop.DBus',
                                        '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')

        #self.ck_manager.connect_to_signal("SeatAdded", self.__seat_added, dbus_interface='org.freedesktop.ConsoleKit.Manager')

        self.nxprovider = provider.Provider(self, self.bus)
        self.enforce_polkit = False
        logger.debug("Started")
 
    # Code for ConsoleKit Multi-seat branch
    #def __seat_added(self, seat_path):
    #    seat_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', seat_path)
    #    seat = dbus.Interface (seat_obj, 'org.freedesktop.ConsoleKit.Seat')
    #    seat.connect_to_signal("SessionToAdd", self.__session_to_add, dbus_interface='org.freedesktop.ConsoleKit.Seat')
    #    seat.ManageSeat()
    #    self.managed_seats[seat.GetId()] = seat
    #    logger.info('Seat %s added' % seat_path)

    #def __session_to_add(self, session_path, is_dynamic, type, parameters, display_type, display_parameters):
    #    logger.info("Session %s to add" % session_path)
    #    if display_type == 'NX':
    #        session_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', session_path)
    #        session = dbus.Interface (session_obj, 'org.freedesktop.ConsoleKit.Seat')
    #        self.__create_session(None, effective_user='', ck_session=session, display_parameters=display_parameters)

    def run(self):
        try:
            logger.debug("Run the main loop")
            self.loop.run()
            logger.debug("Exit the main loop")
        except KeyboardInterrupt:
            logger.debug("keyboardInterrupt")

    def __receive_session_closed_signal(self, session_id):
        logger.debug("SessionClosed received: %s" % session_id)
        if session_id in self.manager_sessions:
            session = self.manager_sessions.pop(session_id)
            session.remove_from_connection()

    def _get_session(self, session_id, type=None):
        logger.debug("Getting session %s" % session_id)
        if session_id not in self.manager_sessions:
            logger.debug ("Session %s not found on db" % session_id)
            logger.debug (self.manager_sessions.keys())
        return self.manager_sessions[session_id]

    def _get_session_for_session_worker (self, session_worker_conn):
        for session in self.manager_sessions.values():
            if session.session_worker_conn == session_worker_conn:
                return session
        return None

    def _get_sessions_for_type(self, type):
        sessions = []
        for session in self.manager_sessions.values():
            if session.seat_type == type:
                sessions.append (session)
        return sessions

    def __create_session(self, type, sender, effective_user=None, ck_session=None, display_parameters=None):
        nxsession = None
        nxsession = Session (sender, type, self.bus, self, self.ck_manager, self.dbus_info, logger, effective_user=effective_user, provider=self.nxprovider)
        self.manager_sessions[nxsession.ck_session_id] = nxsession

        object_path = nxsession.get_object_path()
        logger.debug (object_path.__repr__())
        session_obj = self.bus.get_object (BUS_NAME, object_path)
        session = dbus.Interface (session_obj, SESSION_INTERFACE)
        session.connect_to_signal("SessionClosed", self.__receive_session_closed_signal, dbus_interface=SESSION_INTERFACE)
        return nxsession


    #### DBUS Methods/Signals
    @dbus.service.method( dbus_interface=MANAGER_INTERFACE,
                       in_signature='s', out_signature='o', sender_keyword='sender', connection_keyword='conn' )
    def GetSession( self, session_id, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/SessionManager \
          org.tacix.SessionManager.GetSession \
          string:"user"
        '''
        return self._get_session(session_id).get_object_path()

    @dbus.service.method( dbus_interface=MANAGER_INTERFACE,
                       in_signature='', out_signature='ao', sender_keyword='sender', connection_keyword='conn' )
    def GetSessions( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/SessionManager \
          org.tacix.SessionManager.GetSessions
        '''
        logger.debug ("Getting sessions")
        ops = []
        for session in self.manager_sessions.values():
            ops.append(session.get_object_path())
        return ops

    @dbus.service.method( dbus_interface=MANAGER_INTERFACE,
                       in_signature='s', out_signature='ao', sender_keyword='sender', connection_keyword='conn' )
    def GetSessionsForType( self, type, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/SessionManager \
          org.tacix.SessionManager.GetSessions
        '''
        logger.debug ("Getting sessions for type %s" % type)
        sessions = []
        for session in self._get_sessions_for_type(type):
            sessions.append(session.get_object_path())
        return sessions

    @dbus.service.method( dbus_interface=MANAGER_INTERFACE,
                       in_signature='ss', out_signature='o', sender_keyword='sender', connection_keyword='conn' )
    def CreateSession( self, user, type, sender, conn ):
        '''
        Create a session for the user.

        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/SessionManager \
          org.tacix.SessionManager.CreateSession \
          string:"user"
        '''
        if type == "guest":
            self._check_polkit_privilege(sender, conn, 'org.tacix.sessionmanager.createsession.guest')

        logger.debug ("Creating a new session")
        nxsession = self.__create_session(type, sender, effective_user=user)
        object_path = nxsession.get_object_path()

        return object_path


    def _check_polkit_privilege(self, sender, conn, privilege):
        '''Verify that sender has a given PolicyKit privilege.

        sender is the sender's (private) D-BUS name, such as ":1:42"
        (sender_keyword in @dbus.service.methods). conn is
        the dbus.Connection object (connection_keyword in
        @dbus.service.methods). privilege is the PolicyKit privilege string.

        This method returns if the caller is privileged, and otherwise throws a
        PermissionDeniedByPolicy exception.
        '''
        if sender is None and conn is None:
            # called locally, not through D-BUS: allowing
            return
        if not self.enforce_polkit:
            # that happens for testing purposes when running on the session
            # bus, and it does not make sense to restrict operations here
            return

        # get peer PID
        if self.dbus_info is None:
            self.dbus_info = dbus.Interface(conn.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)
        
        # query PolicyKit
        if self.polkit is None:
            self.polkit = dbus.Interface(dbus.SystemBus().get_object(
                'org.freedesktop.PolicyKit1',
                '/org/freedesktop/PolicyKit1/Authority', False),
                'org.freedesktop.PolicyKit1.Authority')
        try:
            # we don't need is_challenge return here, since we call with AllowUserInteraction
            (is_auth, _, details) = self.polkit.CheckAuthorization(
                    ('unix-process', {'pid': dbus.UInt32(pid, variant_level=1)}), 
                    privilege, {'': ''}, dbus.UInt32(1), '', timeout=600)
        except dbus.DBusException, e:
            if e._dbus_error_name == 'org.freedesktop.DBus.Error.ServiceUnknown':
                # polkitd timed out, connect again
                self.polkit = None
                return self._check_polkit_privilege(sender, conn, privilege)
            else:
                raise

        if not is_auth:
            logging.debug('_check_polkit_privilege: sender %s on connection %s pid %i is not authorized for %s: %s' %
                    (sender, conn, pid, privilege, str(details)))
            raise PermissionDeniedByPolicy(privilege)
