#!/usr/bin/python
#
# Copyright 2007 Google Inc.
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: diamond@google.com (Stephen Shirley)
#         marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import signal
import subprocess
import sys
import gobject

from tacix.util import *
from tacix.server.nx.apps import AppProgram

class AgentException(dbus.DBusException):
    _dbus_error_name = 'org.tacix.session.CKException'

class Agent(AppProgram):
    def __init__( self, nxsession ):
        self.options = None 
        self.txpath = nxsession.txpath
        self.txsession = nxsession.txsession
        self.nxsession = nxsession
        self.logger = self.txpath.get_logger("nxagent")
        self.app_process = None
        self.counter = None
        AppProgram.__init__(self, self.logger)

    # Utilities methods
    def send_signal(self, num):
        try:
            self.logger.info('Send signal %d to agent %d' % (num, self.app_process.pid))
            os.kill(self.app_process.pid, num)
            return True
        except OSError, e:
            return False

    def kill_watchdog(self):
        if self.nxsession.current_state != 'waiting_watchdog':
            return False
        try:
            os.kill(int(self.nxsession.watchdog_pid), signal.SIGTERM)
        except OSError, (errno, strerror):
            self.logger.warn("matched info kill_watchdog, "
                        "got error from kill[%d]: %s\n" % (errno, strerror))
        except Exception, e:
            self.logger.error(e)
        else:
            self.logger.info("matched info kill_watchdog, sent TERM.\n")
        return False


    # Options manipulations functions
    def resume(self, options):
        self.options = options
        self.logger.info("Rewriting the options file")
        self.write_options()

        state = self.nxsession.current_state

        if state == 'running':
            self.suspend()
        elif state == 'suspended':
            self.restore()

    def write_options(self):
        opts = ['nx/nx']

        for opt_name in ("keyboard", "id", "clipboard", "geometry",
                "option", "shadow", "shadowmode", "streaming", "defer", "tile",
                "R", "render", "fullscreen", "shpix", "shmem", "link",
                "media", "cups", "samba", "cache", "images",
                "resize", "backingstore", "menu", "shadowuid", "clients",
                "client", "copysize", "cookie", "cleanup" ):
            if opt_name in self.options:
                opts.append('%s=%s' % (opt_name, self.options[opt_name]))
        opts.append("type=" + self.options['type'].replace('unix-', ''))

        open( self.txpath.get_filepath( "options" ), "w").write("%s:%s\n" % (",".join(opts), self.options['display']))


    # Action methods
    def terminate(self):
        self.send_signal(signal.SIGTERM)

    def suspend(self):
        self.send_signal(signal.SIGHUP)

    def restore(self):
        self.send_signal(signal.SIGHUP)

    def run(self, options):
        self.options = options
        self.logger.info("Starting session %s\n" % self.options['id'])
        self.logger.debug("Exporting DISPLAY for nxagent:")
        display_tmp = 'nx/nx,options=%s:%s' % (self.txpath.get_filepath( "options" ), self.options['display'])
        os.environ['DISPLAY'] = display_tmp

        self.logger.debug("Running xauth")
        if not self.setup_xauth():
            self.logger.debug("xauth returned")
            self.nxsession.change_state('denied')
            return False

        if not self.start_agent():
            return False

        # Revert DISPLAY env for the app
        os.environ['DISPLAY'] = ':%s' % self.options['display']
        return True

    def start_agent(self):
        self.logger.info("Starting nxagent\n")
        self.write_options()
        self.logger.info("Options wrote\n")
        cmd = []
        cmd.append('/usr/bin/nxagent')
        if self.options['seat-type'] == 'nx-shadow':
            cmd.append('-S')
            cmd.append('-shadow')
            #cmd.append('%(shadowhost)s:%(shadowdisplay)s' % self.options)
            cmd.append(':%(shadowdisplay)s' % self.options)
            cmd.append('-shadowmode')
            cmd.append('1') #Interactive showdowing [1|0]
        else:
            if self.options['rootless'] == '1': cmd.append('-R')
            else: cmd.append('-D')

            if self.options['persistent'] == '1': cmd.append('-persistent')
            else: cmd.append('-nopersistent')

        cmd.append('-name')
        if 'session' in self.options:
            cmd.append('"TaciX - %s@%s:%s - %s" ' % \
                         (self.options['user'], os.uname()[1], self.options['display'], self.options['session']))
        else:
            cmd.append('"TaciX - %s@%s:%s - %s" ' % \
                         (self.options['user'], os.uname()[1], self.options['display'], 'session'))
        cmd.append(':%s' % self.options['display'])

        self.logger.debug("Agent cmd: %s\n" % cmd)

        self.app_process = subprocess.Popen(cmd,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if self.app_process is None:
            self.logger.info("Could not execute nxagent \n")
            return False
        self._child_add_watch()
        self.logger.info("Spawned nxagent pid %d\n" % self.app_process.pid)
        return True

    def setup_xauth(self):
        """Setup Xauthority file using session cookie."""
        authority_file = self.options['authority_file']

        if self.options['seat-type'] == 'nx-shadow':
            os.environ['SHADOW_XAUTHORITY'] = self.options['shadow_authority_file']
            os.environ['XAUTHORITY'] = authority_file
        else:
            os.environ['XAUTHORITY'] = authority_file
        return True


