#!/usr/bin/python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import subprocess
import gobject
import signal

class AppProgram:
    def __init__(self, logger, cmd=None, log_path=None, callback=None):
        self.app_process = None
        self.app_watcher = None

        self.cmd = cmd
        self.log_path = log_path
        self.logger = logger
        self.callback = callback

    def terminate(self):
        self._send_signal(signal.SIGTERM)

    def _run(self):
        log_file = open(self.log_path, 'w')
        self.app_process = subprocess.Popen(self.cmd, stdout=log_file, stderr=log_file, close_fds=True)
        if self.callback is not None:
            self._child_add_watch()
        self.logger.debug("Subprocess wih pid: %r" % self.app_process.pid)

    def _child_add_watch(self):
        if self.app_process.pid is None:
            return
        if self.app_watcher is not None:
            return
        self.app_watcher = gobject.child_watch_add(self.app_process.pid, self._on_app_end)
        self.logger.debug("App watcher added")

    def _on_app_end(self, pid, error_code):
        self.logger.debug('App terminated')
        if self.callback is not None:
            self.callback()

    def _send_signal(self, num):
        if self.app_process is None:
            return False
        try:
            self.logger.info('Send signal %d to agent %d' % (num, self.app_process.pid))
            os.kill(self.app_process.pid, num)
            return True
        except OSError, e:
            return False


class XSessionProgram(AppProgram):
    def __init__(self, nxagent, logger, txpath):
        cmd = ["/etc/tacix/Xsession"]
        self.nxagent = nxagent
        AppProgram.__init__(self, logger, cmd=cmd, log_path=txpath.get_filepath('app.log'), callback=self.nxagent.terminate)

    def _run(self):
        """Start the session application (startkde/gnome-session/etc)."""
        self.logger.info("NXSessionApp ready, launching app")

        os.environ['NX_SESSION_TYPE'] = self.nxagent.options['type']
        if 'application' in self.nxagent.options:
            os.environ['NX_APPLICATION'] = self.nxagent.options['application']
        AppProgram._run(self)


class VncProgram(AppProgram):
    def __init__(self, nxagent, logger, txpath):
        '''
        --type="vnc" --agent_server="host%3A1234" --agent_password="password"
        '''
        self.logger.info("NXSessionApp ready, launching vncviewer")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        cmd = ['vnc']
        AppProgram.__init__(self, logger, cmd=cmd, log_path=txpath.get_filepath('app.log'), callback=self.nxagent.terminate)

    def _run(self):
        AppProgram._run(self)


class RDesktopProgram(AppProgram):
    def __init__(self, nxagent, logger, txpath):
        cmd = ['rdesktop',
                '-u', nxagent.options['agent_user'],
                '-k', nxagent.options['agent_keyboard']]
        if 'agent_domain' in nxagent.options and nxagent.options['agent_domain'] != '':
            cmd.append('-d')
            cmd.append(nxagent.options['agent_domain'])
        if 'application' in nxagent.options and nxagent.options['application'] != '':
            cmd.append('-s')
            cmd.append(nxagent.options['application'])
        cmd.append('-f')
        cmd.append(nxagent.options['agent_server'])
        AppProgram.__init__(self, logger, cmd=cmd, log_path=txpath.get_filepath('app.log'), callback=self.nxagent.terminate)

    def _run(self):
        log_file = open(self.log_path, 'w')
        self.app_process = subprocess.Popen(self.cmd, stdin=subprocess.PIPE, stdout=log_file, stderr=log_file, close_fds=True)
        self.logger.debug("Subprocess wih pid: %r" % self.app_process.pid)
        self.app_process.stdin.write(self.nxagent.options['agent_password'])
        self.app_process.stdin.close()

class SessionRunner(AppProgram):
    def __init__(self, users):
        self.users = users

    def run(self):
        os.environ['NX_USERS'] = this.users.join(',')
        self._run()
        self._child_add_watch()
        del os.environ['NX_USERS']

class ShadowDialog(AppProgram):
    def __init__(self, setup_func, callback, logger):
        self.callback = callback
        self.setup_func = setup_func
        cmd = ['']
        AppProgram.__init__(self, cmd=cmd)

    def run(self):
        self.setup_func()
        self._run()
        self._child_add_watch()

    def _on_app_end(self, pid, error_code):
        AppProgram._on_app_end(self, pid, error_code)
        self.callback(error_code)

