#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import dbus
import dbus.service
import pwd

import gobject

from tacix.server.nx.nxutils import *
from tacix import util
from tacix.server.nx.apps import SessionRunner
from tacix.server.nx.config import Config
from tacix.server.nx.display import *

logger = util.get_logger('txmanager', '/var/log/tacixmanager.log')

PROVIDER_BUS_NAME='org.tacix.nx'
PROVIDER_PATH='/org/tacix/nx/provider'
PROVIDER_INTERFACE='org.tacix.nx.provider'

class Provider( dbus.service.Object ):
    def __init__( self, tacix_manager, bus ):
        logger.debug("Starting NXProvider")
        #from dbus.mainloop.glib import DBusGMainLoop 
 	#DBusGMainLoop(set_as_default=True)

        # Getting the system dbus
        #self.bus = dbus.SystemBus( )
        self.bus = bus
        
        bus_name = dbus.service.BusName(PROVIDER_BUS_NAME,
                                        self.bus,
                                        do_not_queue=True)
        dbus.service.Object.__init__(self, bus_name=bus_name, object_path=PROVIDER_PATH)
        #self.loop = gobject.MainLoop()

        self.ck_manager = dbus.Interface (
                self.bus.get_object ('org.freedesktop.ConsoleKit', '/org/freedesktop/ConsoleKit/Manager'),
                'org.freedesktop.ConsoleKit.Manager')

        #self.dbus_info = dbus.Interface(self.bus.get_object('org.freedesktop.DBus',
        #                                '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')

        self.tacix_manager = tacix_manager
        logger.debug("NXProvider Started")
 
        self.nxconfig = Config()

    def __get_resumable_sessions(self, user, options):
        sessions = {}
        for session in self.tacix_manager._get_sessions_for_type('nx'):
            properties = session.get_properties()
            if 'persistent' in properties and properties['persistent'] == '0':
                continue
            if properties['user'] != user:
                continue
            if 'status' in options and options['status'].find(properties['state']) == -1:
                continue
            logger.debug("Adding session %s" % properties)
            sessions[session.get_property('display')] = properties
        return sessions

    def __start_sessions(self, users):
        server = SessionRunner(users)
        return server.run()

    def is_allowed( self, property, value ):
        handler_name = '_property_%s_handler' % property
        try:
            handler_method = getattr(self, handler_name)
            ret = handler_method(property, value)
            return ret
        except AttributeError, e:
            pass

        ret = self.nxconfig.get_allowed_value(property, value)
        if ret is None:
            return None
        return {property: ret}

    def _property_screeninfo_handler(self, prop, value):
        split = value.split('x')
        return {'depth': split[2].split('+')[0], 'screen': "%sx%s" % (split[0], split[1])}

    def _property_id_handler(self, prop, value):
        return None

    def find_next_available_display(self):
        return str(find_next_available_display(self.nxconfig.get_default_option('display')))

    def get_default_options(self):
        return self.nxconfig.get_default_options()

    def calculate_options(self, uid, display, ck_session_id):
        '''
           Calculate some options
        '''
        options = {
            'display' : display,
            'hostname' : os.uname()[1],
            'state' : 'none',
            'cli_opts': '-RD--PS-', # FRD--PSA Used by listsessions
            'available': 'Yes', # N/A Used by listsessions
            'userip' : '127.0.0.1',
            'proxyip' : '127.0.0.1',
            'proxyip' : '127.0.0.1',
            'host' : '127.0.0.1',
            'uid' : uid,
            'user': pwd.getpwuid(int(uid)).pw_name,
            'ck_session_id': ck_session_id
            }

        options['id'] = "%s-%s-%s" % (options['hostname'], display, ck_session_id)

        for opt_name in options:
            options[opt_name] = self.nxconfig.get_allowed_value( opt_name, options[opt_name] )

        return options


    #### DBUS Methods/Signals
    def FindNextAvailableDisplay (self):
        return self.find_next_available_display()

    def GetDefaultOptions (self):
        return self.get_default_options()

    def CalculateOptions (self, uid, display, ck_session_id):
        return self.calculate_options(uid, display, ck_session_id)

    def IsAllowed (self, property, value):
        return self.is_allowed (property, value)

    @dbus.service.method( dbus_interface=PROVIDER_INTERFACE,
                       in_signature='s', out_signature='a{ss}', sender_keyword='sender', connection_keyword='conn' )
    def GetNXProxyOptions( self, session_id, sender, conn ):
        '''
        Gets everything needed by nxproxy to run.
        This method is or test porpuse, invoking should be restricted.

        Example:
        dbus-send --system --dest=org.tacix.nx \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/nx/provider \
          org.tacix.nx.provider \
          string:"user"
        '''
        properties = self.tacix_manager._get_session(session_id).get_properties_complete()
        ret = parse_parameters_dict(properties)
        return ret
 
    @dbus.service.method( dbus_interface=PROVIDER_INTERFACE,
                       in_signature='as', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def StartSessions( self, users, sender, conn ):
        '''
        Creates a new session and starts it on a hidden X server.
        This method is or test porpuse, invoking should be restricted.

        Example:
        dbus-send --system --dest=org.tacix.nx \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/nx/provider \
          org.tacix.nx.provider \
          string:"user"
        '''
        return self.__start_session(users)

    @dbus.service.method( dbus_interface=PROVIDER_INTERFACE,
                       in_signature='a{ss}', out_signature='a{sa{ss}}', sender_keyword='sender', connection_keyword='conn' )
    def GetResumableSessions(self, options, sender=None, conn=None):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/nx/provider \
          org.tacix.nx.provider.GetResumableSessions
        '''
        #return {}
        #self._check_polkit_privilege(sender, conn, 'org.tacix.sessionmanager.resume')
        return self.__get_resumable_sessions(options['user'], options)

    @dbus.service.method( dbus_interface=PROVIDER_INTERFACE,
                       in_signature='', out_signature='a{sa{ss}}', sender_keyword='sender', connection_keyword='conn' )
    def GetShadowableSessions(self, sender=None, conn=None):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/nx/provider \
          org.tacix.nx.provider.GetShadowableSessions
        '''

        #self._check_polkit_privilege(sender, conn, 'org.tacix.sessionmanager.resume')
        sessions = {}
        ck_sessions = self.ck_manager.GetSessions()
        for ck_op_session in ck_sessions:
            ck_session = dbus.Interface (self.bus.get_object ('org.freedesktop.ConsoleKit', ck_op_session), 'org.freedesktop.ConsoleKit.Session')
            display = ck_session.GetX11Display().replace(':', '')
            if display != '' and int(display) < 100:
                session_opts = {'display': display,
                    'type': 'console',
                    'cli_opts': '--------',
                    'depth': '',
                    'screen': 'x',
                    'available': 'Running',
                    'state': 'Running',
                    'session': 'X%s (Local)' % display,
                    'ck_session_id': ck_session.GetId().rsplit('/', 2)[-1]}
                sessions[session_opts['display']] = session_opts

        for session in self.tacix_manager.manager_sessions.values():
            props = session.get_properties()
            props['available'] = 'Yes'
            props['ck_session_id'] = session.ck_session_id
            props['session'] = session.ck_session_id
            sessions[session.get_property('display')] = props
        return sessions

    @dbus.service.method( dbus_interface=PROVIDER_INTERFACE,
                       in_signature='s', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def CreateShadowAuthority(self, ck_session_id, sender, conn):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.AskShadowPermission
        '''
        logger.debug('Asking for shadow permission')
        session = self.tacix_manager._get_session(ck_session_id)
        #dialog = ShadowDialog (setup_func, callback, self.logger)
        #dialog.run()
        authority_file = session.AddUserAuthorization('nx')

        shadow_session = self.tacix_manager._get_session_for_session_worker(sender)
        shadow_session.set_property('shadowdisplay', session.get_property('display'))
        shadow_session.set_property('shadow_authority_file', authority_file)

        #session_worker = dbus.Interface (self.bus.get_object (sender, '/Session'), 'org.tacix.session.worker')
        #session_worker.SetShadowAuthority(authority_file)

    @dbus.service.method( dbus_interface=PROVIDER_INTERFACE,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def CreateShadow(self, user, sender, conn):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.CreateShadow
        '''
        file = xauth.create_xauthority_file_for_user(user, self.get_property('cookie'), self.get_property('display'))
        if file is None:
            raise TaciXSessionException('Could not create the authority file')
        return file
 

