#!/usr/bin/python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import os.path
import signal
import subprocess
import traceback
import gobject
import shutil

import glib
import dbus
import dbus.service
import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

from tacix.server.nx.agent import Agent
from tacix.server.nx.agentmonitor import AgentMonitor
from tacix.server.nx.apps import *
#from tacix.client.resources.TXResources import TXResources
from tacix.util import *
from tacix.constants import *

class SessionWorkerException(dbus.DBusException):
    _dbus_error_name = 'org.tacix.session.SessionWorkerException'

class SessionWorker(dbus.service.Object):
    def __init__ ( self, sessionid ):
        self.bus = dbus.SystemBus()
        dbus.service.Object.__init__(self, self.bus, "/Session")
        self.loop = gobject.MainLoop()

        signal.signal(signal.SIGHUP, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

        self.txsession = get_session(sessionid, bus=self.bus)
        if not self.txsession.RegisterSessionWorker(''):
            return

        self.txpath = TXPath(sessionid, create=True)
        self.logger = self.txpath.get_logger("nxsession")

        self.seat_type = None
        self.nxagent = Agent(self)
        self.nxagentmonitor = AgentMonitor(self)
        self.app_program = None
        self.app_signal = None

        #self.txresources = TXResources(sessionid)
        self.current_state = None
        self.watchdog_pid = None
        self.action = None

        self.change_state('ready')

    # Watchers
    def signal_handler(self, num, stack):
        self.logger.info('Received signal %d' % num)
        if self.nxagent != None:
            self.logger.info('Sending signal to agent')
            self.nxagent.send_signal(num)

    def change_state(self, new_state):
        if self.current_state == new_state:
            self.logger.info("Failed to change the agent status. Status already is %s" % new_state)
            return

        old_state = self.current_state
        self.current_state = new_state

        self.logger.info("App watcher state is %s" % new_state)
        if self.app_program is None and self.seat_type == 'nx' and new_state == 'running':
            self.run_app()
        elif new_state == 'terminated':
            # Add a timeout of 2s to cleanup
            gobject.timeout_add(2000, self.cleanup)
        elif new_state == 'waiting_watchdog':
            # Add a timeout of 5s to kill the watchdog
            #gobject.timeout_add(1000, self.nxagent.kill_watchdog)
            self.nxagent.kill_watchdog()
        elif new_state == 'failed' and old_state == 'running':
            self.logger.info("Seems that the failed state is for a running session, so suspending")
        elif new_state == 'suspended' and self.action == 'resume':
            self.nxagent.restore()
        elif new_state == 'running':
            self.action = None

        self.logger.debug("Setting dbus txsession state: %s" % new_state)
        self.StateChanged(new_state)
        self.logger.info("Nxagent state was: %s Now: %s\n" % \
                (old_state, new_state))

    # Signals
    @dbus.service.signal( SESSION_WORKER_INTERFACE )
    def StateChanged (self, state):
        self.logger.debug( "StateChanged emitted, state = %s" % state )

    # Agent actions
    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='', out_signature='s')
    def InterfaceVersion( self ):
        return '1'

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='', out_signature='s')
    def GetStatus( self ):
        return self.current_state

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='sa{ss}', out_signature='b')
    def Run( self, ck_cookie, options ):
        if ck_cookie == '':
            raise SessionWorkerException('Cookie not provided')
        self.seat_type = options['seat-type']

        os.environ['XDG_SESSION_COOKIE'] = ck_cookie
        os.environ['NX_CLIENT'] = '/usr/NX/bin/nxclient'

        self.logger.debug("TXManager session openned\n")

        # Add watchers for the sessionapp
        #self.txresources.setup()

        # Run the agent
        if not self.nxagent.run(options):
            return False
        self.nxagentmonitor.run()
        return True

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='a{ss}', out_signature='b')
    def Resume(self, options):
        if self.nxagent is None:
            return False
        self.action = 'resume'
        self.nxagent.resume(options)
        return True

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='', out_signature='b')
    def Suspend (self):
        self.nxagent.suspend()
        return True

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='', out_signature='b')
    def Terminate (self):
        self.nxagent.terminate()
        return True

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='s', out_signature='')
    def SetShadowAuthority (self, authority_file):
        self.shadow_authority_file = authority_file

    def wait( self ):
        self.loop.run()

    # App actions
    def run_app(self):
        if self.nxagent.options['seat-type'] == 'nx-shadow':
            return
        type = self.nxagent.options['type']
        if type == 'windows':
            self.app_program = RDesktopProgram(self.nxagent, self.logger, self.txpath)
        elif type == 'vnc':
            self.app_program = VncProgram(self.nxagent, self.logger, self.txpath)
        else:
            self.app_program = XSessionProgram(self.nxagent, self.logger, self.txpath)
        self.app_program._run()
        self.app_program._child_add_watch()

    def cleanup(self, status=None, keep_log=True):
        """Cleanup after various things (X11 locks etc)."""

        self.logger.info('Cleaning up session\n')

        for file in ('/tmp/.X%(display)s-lock' % self.nxagent.options,
                  '/tmp/.nX%(display)s-lock' % self.nxagent.options,
                  '/tmp/.X11-unix/X%(display)s' % self.nxagent.options):
            if os.path.exists(file):
                try:
                    os.remove(file)
                except OSError, e:
                    self.logger.error("Error removing file. %s\n" % e)

        # Move if failed or terminated if defined
        self.logger.info('NXSession Exited')
        if not keep_log:
            shutil.rmtree(self.txpath.get_dir())
        elif status is None:
            shutil.move(self.txpath.get_dir(), self.txpath.get_unknown_dir())
        elif status == 'terminated':
            shutil.move(self.txpath.get_dir(), self.txpath.get_terminated_dir())
        elif status == 'failed':
            shutil.move(self.txpath.get_dir(), self.txpath.get_failed_dir())
        self.loop.quit()
        return False


class ShadowSessionWorker (SessionWorker):
    #@dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
    #                     in_signature='sa{ss}', out_signature='b')
    #def Run( self, ck_cookie, options ):
    #    #provider = dbus.Interface (bus.get_object (PROVIDER_BUS_NAME, PROVIDER_PATH), PROVIDER_INTERFACE)
    #    if options['shadow_authority_file'] in (None, ''):
    #        raise SessionWorkerException('Not authorized')
    #    SessionWorker.Run(self, ck_cookie, options)

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='a{ss}', out_signature='b')
    def Resume(self, options):
        raise SessionWorkerException('Action not supported by shadow session')

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='', out_signature='b')
    def Suspend (self):
        raise SessionWorkerException('Action not supported by shadow session')

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='s', out_signature='')
    def RunShadowForCKSession (self, ck_session_id):
        provider = dbus.Interface (self.bus.get_object (PROVIDER_BUS_NAME, PROVIDER_PATH), PROVIDER_INTERFACE)
        provider.CreateShadowAuthority (ck_session_id)

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='s', out_signature='')
    def RunShadowForDisplay (self, display):
        provider = dbus.Interface (self.bus.get_object (PROVIDER_BUS_NAME, PROVIDER_PATH), PROVIDER_INTERFACE)
        ck_session_id = get_ck_session_for_display (self.bus, display)
        provider.CreateShadowAuthority (ck_session_id)

    @dbus.service.method(dbus_interface=SESSION_WORKER_INTERFACE,
                         in_signature='', out_signature='b')
    def Terminate (self):
        self.nxagent.terminate()
        return True


def get_ck_session_for_display (bus, display):
    # Getting the ConsoleKit object
    ck_manager = dbus.Interface (
            bus.get_object ('org.freedesktop.ConsoleKit', '/org/freedesktop/ConsoleKit/Manager'),
            'org.freedesktop.ConsoleKit.Manager')

    ck_session_id = None
    ck_sessions = ck_manager.GetSessions()
    for ck_op_session in ck_sessions:
        ck_session = dbus.Interface (bus.get_object ('org.freedesktop.ConsoleKit', ck_op_session), 'org.freedesktop.ConsoleKit.Session')
        if ck_session.GetX11Display() == display:
            return ck_session.GetId ()
    return None
    #self.logger.debug ('Starting shadow for ck ' + ck_session_id)

